import inspect
import os
import subprocess

from dataclasses import dataclass
from datetime import datetime
from email.message import Message
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from mimetypes import guess_type
from typing import Optional, List, Union, Any, Dict

from platypush.message import JSONAble
from platypush.plugins import Plugin, action


@dataclass
class ServerInfo:
    server: str
    port: int
    username: Optional[str]
    password: Optional[str]
    ssl: bool
    keyfile: Optional[str]
    certfile: Optional[str]
    access_token: Optional[str]
    oauth_mechanism: Optional[str]
    oauth_vendor: Optional[str]
    timeout: Optional[int]


class Mail(JSONAble):
    def __init__(self, id: int, date: datetime, size: int,
                 from_: Optional[Union[Dict[str, str], List[str]]] = None,
                 to: Optional[Union[Dict[str, str], List[str]]] = None,
                 cc: Optional[Union[Dict[str, str], List[str]]] = None,
                 bcc: Optional[Union[Dict[str, str], List[str]]] = None, subject: str = '',
                 payload: Optional[Any] = None, **kwargs):
        self.id = id
        self.date = date
        self.size = size
        self.from_ = from_ or kwargs.get('from')
        self.to = to
        self.cc = cc or []
        self.bcc = bcc or []
        self.subject = subject
        self.payload = payload

        for k, v in kwargs.items():
            setattr(self, k, v)

    def to_json(self) -> dict:
        return {
            k if k != 'from_' else 'from': v
            for k, v in dict(inspect.getmembers(self)).items()
            if not k.startswith('_') and not callable(v)
        }


class MailPlugin(Plugin):
    """
    Base class for mail plugins.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.server_info: Optional[ServerInfo] = None

    @staticmethod
    def _get_password(password: Optional[str] = None, password_cmd: Optional[str] = None) -> Optional[str]:
        """
        Get the password either from a provided string or from a password command.
        """
        if not password_cmd:
            return password

        proc = subprocess.Popen(['sh', '-c', password_cmd], stdout=subprocess.PIPE)
        password = proc.communicate()[0].decode()
        return password or None

    @staticmethod
    def _get_path(path: str) -> str:
        return os.path.abspath(os.path.expanduser(path))

    def _get_server_info(self, server: Optional[str] = None, port: Optional[int] = None, username: Optional[str] = None,
                         password: Optional[str] = None, password_cmd: Optional[str] = None,
                         ssl: Optional[bool] = None, keyfile: Optional[str] = None, certfile: Optional[str] = None,
                         access_token: Optional[str] = None, oauth_mechanism: Optional[str] = None,
                         oauth_vendor: Optional[str] = None, default_port: Optional[int] = None,
                         default_ssl_port: Optional[int] = None, timeout: Optional[int] = None, **kwargs) \
            -> ServerInfo:
        if not port:
            port = default_ssl_port if ssl else default_port

        info = ServerInfo(server=server, port=port, username=username,
                          password=self._get_password(password, password_cmd), ssl=ssl, keyfile=keyfile,
                          certfile=certfile, access_token=access_token, oauth_mechanism=oauth_mechanism,
                          oauth_vendor=oauth_vendor, timeout=timeout)

        if server:
            return info

        if self.server_info:
            assert self.server_info.server, 'No server specified'
            return self.server_info

        return info


class MailInPlugin(MailPlugin):
    """
    Base class for mail in plugins.
    """

    @action
    def get_folders(self) -> list:
        raise NotImplementedError()

    @action
    def get_sub_folders(self) -> list:
        raise NotImplementedError()

    @action
    def search(self, criteria: str, directory: Optional[str] = None) -> list:
        raise NotImplementedError()

    @action
    def search_unseen_messages(self, directory: Optional[str] = None) -> list:
        raise NotImplementedError()

    def search_flagged_messages(self, folder: str = 'INBOX', **connect_args) -> list:
        raise NotImplementedError()

    @action
    def get_message(self, id) -> dict:
        raise NotImplementedError()


class MailOutPlugin(MailPlugin):
    """
    Base class for mail out plugins.
    """

    def send_message(self, message: Message, **connect_args):
        raise NotImplementedError()

    @staticmethod
    def _file_to_part(file: str) -> MIMEBase:
        _type, _subtype, _type_class = 'application', 'octet-stream', MIMEApplication
        mime_type, _sub_subtype = guess_type(file)

        if mime_type:
            _type, _subtype = mime_type.split('/')
        if _sub_subtype:
            _subtype += ';' + _sub_subtype

        if _type == 'application':
            _type_class = MIMEApplication
        elif _type == 'audio':
            _type_class = MIMEAudio
        elif _type == 'image':
            _type_class = MIMEImage
        elif _type == 'text':
            _type_class = MIMEText

        args = {}
        if _type_class != MIMEText:
            mode = 'rb'
            args['Name'] = os.path.basename(file)
        else:
            mode = 'r'

        with open(file, mode) as f:
            return _type_class(f.read(), _subtype, **args)

    @classmethod
    def create_message(cls, to: Union[str, List[str]], from_: Optional[str] = None,
                       cc: Optional[Union[str, List[str]]] = None, bcc: Optional[Union[str, List[str]]] = None,
                       subject: str = '', body: str = '', body_type: str = 'plain',
                       attachments: Optional[List[str]] = None, headers: Optional[Dict[str, str]] = None) -> Message:
        assert from_, 'from/from_ field not specified'

        body = MIMEText(body, body_type)
        if attachments:
            msg = MIMEMultipart()
            msg.attach(body)

            for attachment in attachments:
                attachment = os.path.abspath(os.path.expanduser(attachment))
                assert os.path.isfile(attachment), 'No such file: {}'.format(attachment)
                part = cls._file_to_part(attachment)
                part['Content-Disposition'] = 'attachment; filename="{}"'.format(os.path.basename(attachment))
                msg.attach(part)
        else:
            msg = body

        msg['From'] = from_
        msg['To'] = ', '.join(to) if isinstance(to, List) else to
        msg['Cc'] = ', '.join(cc) if cc else ''
        msg['Bcc'] = ', '.join(bcc) if bcc else ''
        msg['Subject'] = subject

        if headers:
            for name, value in headers.items():
                msg.add_header(name, value)

        return msg

    @action
    def send(self, to: Union[str, List[str]], from_: Optional[str] = None,
             cc: Optional[Union[str, List[str]]] = None, bcc: Optional[Union[str, List[str]]] = None,
             subject: str = '', body: str = '', body_type: str = 'plain', attachments: Optional[List[str]] = None,
             headers: Optional[Dict[str, str]] = None, **connect_args):
        """
        Send an email through the specified SMTP sender.

        :param to: Receiver(s), as comma-separated strings or list.
        :param from_: Sender email address (``from`` is also supported outside of Python contexts).
        :param cc: Carbon-copy addresses, as comma-separated strings or list
        :param bcc: Blind carbon-copy addresses, as comma-separated strings or list
        :param subject: Mail subject.
        :param body: Mail body.
        :param body_type: Mail body type, as a subtype of ``text/`` (e.g. ``html``) (default: ``plain``).
        :param attachments: List of attachment files to send.
        :param headers: Key-value map of headers to be added.
        :param connect_args: Parameters for ``.connect()``, if you want to override the default server configuration.
        """
        if not from_ and 'from' in connect_args:
            from_ = connect_args.pop('from')

        msg = self.create_message(to=to, from_=from_, cc=cc, bcc=bcc, subject=subject, body=body, body_type=body_type,
                                  attachments=attachments, headers=headers)

        return self.send_message(msg, **connect_args)


# vim:sw=4:ts=4:et:



from email.message import Message
from typing import Optional, List

from smtplib import SMTP, SMTP_SSL

from platypush.plugins.mail import MailOutPlugin, ServerInfo


class MailSmtpPlugin(MailOutPlugin):
    """
    Plugin to interact with a mail server over SMTP.
    """

    _default_port = 25
    _default_ssl_port = 465

    def __init__(self, server: Optional[str] = None, port: Optional[int] = None, local_hostname: Optional[str] = None,
                 source_address: Optional[List[str]] = None, username: Optional[str] = None,
                 password: Optional[str] = None, password_cmd: Optional[str] = None, access_token: Optional[str] = None,
                 oauth_mechanism: Optional[str] = 'XOAUTH2', oauth_vendor: Optional[str] = None, ssl: bool = False,
                 keyfile: Optional[str] = None, certfile: Optional[str] = None, timeout: Optional[int] = 60, **kwargs):
        """
        :param server: Server name/address.
        :param port: Port (default: 25 for plain, 465 for SSL).
        :param local_hostname: If specified, local_hostname is used as the FQDN of the local host in the HELO/EHLO
            command. Otherwise, the local hostname is found using socket.getfqdn().
        :param source_address:  The optional source_address parameter allows binding to some specific source address in
            a machine with multiple network interfaces, and/or to some specific source TCP port. It takes a 2-tuple
            (host, port), for the socket to bind to as its source address before connecting. If omitted (or if host or
            port are '' and/or 0 respectively) the OS default behavior will be used.
        :param username: SMTP username.
        :param password: SMTP password.
        :param password_cmd: If you don't want to input your password in the configuration, run this command to fetch
            or decrypt the password.
        :param access_token: OAuth2 access token if the server supports OAuth authentication.
        :param oauth_mechanism: OAuth2 mechanism (default: ``XOAUTH2``).
        :param oauth_vendor: OAuth2 vendor (default: None).
        :param ssl: Use SSL (default: False).
        :param keyfile: Private key file for SSL connection if client authentication is required.
        :param certfile: SSL certificate file or chain.
        :param timeout: Server connect/read timeout in seconds (default: 60).
        """
        super().__init__(**kwargs)
        self.local_hostname = local_hostname
        self.source_address = source_address
        self.server_info = self._get_server_info(server=server, port=port, username=username, password=password,
                                                 password_cmd=password_cmd, ssl=ssl, keyfile=keyfile, certfile=certfile,
                                                 access_token=access_token, oauth_mechanism=oauth_mechanism,
                                                 oauth_vendor=oauth_vendor, timeout=timeout)

    def _get_server_info(self, server: Optional[str] = None, port: Optional[int] = None, username: Optional[str] = None,
                         password: Optional[str] = None, password_cmd: Optional[str] = None,
                         ssl: Optional[bool] = None, keyfile: Optional[str] = None, certfile: Optional[str] = None,
                         timeout: Optional[int] = None, **kwargs) -> ServerInfo:
        return super()._get_server_info(server=server, port=port, username=username, password=password,
                                        password_cmd=password_cmd, ssl=ssl, keyfile=keyfile, certfile=certfile,
                                        default_port=self._default_port, default_ssl_port=self._default_ssl_port,
                                        timeout=timeout)

    def connect(self, **connect_args) -> SMTP:
        info = self._get_server_info(**connect_args)
        self.logger.info('Connecting to {}'.format(info.server))
        smtp_args = {
            'host': info.server,
            'port': info.port,
            'local_hostname': self.local_hostname,
            'source_address': self.source_address,
        }

        if info.ssl:
            client_type = SMTP_SSL
            smtp_args.update(certfile=info.certfile, keyfile=info.keyfile)
        else:
            client_type = SMTP

        client = client_type(**smtp_args)
        if info.password:
            client.login(info.username, info.password)

        return client

    def send_message(self, message: Message, **connect_args):
        with self.connect(**connect_args) as client:
            errors = client.sendmail(message['From'], message['To'], message.as_string())

        if errors:
            return None, ['{}: {}'.format(code, err) for code, err in errors.items()]


# vim:sw=4:ts=4:et:



from concurrent.futures import ThreadPoolExecutor
from typing import Callable


def func_wrapper(args):
    (f, offset, *args) = args
    items = f(*args)
    return [(i + offset, item) for i, item in enumerate(items)]


def get_items(
    func: Callable,
    *args,
    parse: Callable = lambda _: _,
    chunk_size: int = 100,
    processes: int = 5,
):
    """
    This function performs pagination on a function that supports
    `limit`/`offset` parameters and it runs API requests in parallel to speed
    things up.
    """
    items = []
    offsets = [-chunk_size]
    remaining = chunk_size * processes

    with ThreadPoolExecutor(
        processes, thread_name_prefix=f'mopidy-tidal-{func.__name__}-'
    ) as pool:
        while remaining == chunk_size * processes:
            offsets = [offsets[-1] + chunk_size * (i + 1) for i in range(processes)]

            pool_results = pool.map(
                func_wrapper,
                [
                    (
                        func,
                        offset,
                        *args,
                        chunk_size,  # limit
                        offset,  # offset
                    )
                    for offset in offsets
                ],
            )

            new_items = []
            for results in pool_results:
                new_items.extend(results)

            remaining = len(new_items)
            items.extend(new_items)

    items = sorted([_ for _ in items if _], key=lambda item: item[0])
    sorted_items = [item[1] for item in items]
    return list(map(parse, sorted_items))



import json
import uuid

from platypush.message.event import Event
from platypush.utils import get_redis


class WebhookEvent(Event):
    """
    Event triggered when a custom webhook is called.
    """

    def __init__(
        self,
        *argv,
        hook,
        method,
        data=None,
        args=None,
        headers=None,
        response=None,
        **kwargs,
    ):
        """
        :param hook: Name of the invoked web hook, from http://host:port/hook/<hook>
        :type hook: str

        :param method: HTTP method (in uppercase)
        :type method: str

        :param data: Extra data passed over POST/PUT/DELETE
        :type data: str or dict/list from JSON

        :param args: Extra query string arguments
        :type args: dict

        :param headers: Request headers
        :type args: dict

        :param response: Response returned by the hook.
        :type args: dict | list | str
        """
        # This queue is used to synchronize with the hook and wait for its completion
        kwargs['response_queue'] = kwargs.get(
            'response_queue', f'platypush/webhook/{str(uuid.uuid1())}'
        )

        super().__init__(
            *argv,
            hook=hook,
            method=method,
            data=data,
            args=args or {},
            headers=headers or {},
            response=response,
            **kwargs,
        )

    def send_response(self, response):
        output = response.output
        if isinstance(output, tuple):
            # A 3-sized tuple where the second element is an int and the third
            # is a dict represents an HTTP response in the format `(data,
            # http_code headers)`.
            if (
                len(output) == 3
                and isinstance(output[1], int)
                and isinstance(output[2], dict)
            ):
                output = {
                    '___data___': output[0],
                    '___code___': output[1],
                    '___headers___': output[2],
                }
            else:
                # Normalize tuples to lists before serialization
                output = list(output)
        if isinstance(output, (dict, list)):
            output = json.dumps(output)

        if output is None:
            output = ''
        get_redis().rpush(self.args['response_queue'], output)

    def wait_response(self, timeout=None):
        rs = get_redis().blpop(self.args['response_queue'], timeout=timeout)
        if rs and len(rs) > 1:
            return rs[1]


# vim:sw=4:ts=4:et:



from typing import Optional, Dict

from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorDhtPlugin(GpioSensorPlugin):
    """
    Plugin to interact with a DHT11/DHT22/AM2302 temperature/humidity sensor through GPIO.

    Requires:

        * ``Adafruit_Python_DHT`` (``pip install git+https://github.com/adafruit/Adafruit_Python_DHT.git``)

    """

    def __init__(self, sensor_type: str, pin: int, retries: int = 5,
                 retry_seconds: int = 2, **kwargs):
        """
        :param sensor_type: Type of sensor to be used (supported types: DHT11, DHT22, AM2302).
        :param pin: GPIO PIN where the sensor is connected.
        :param retries: Number of retries in case of failed read (default: 5).
        :param retry_seconds: Number of seconds to wait between retries (default: 2).
        """
        super().__init__(**kwargs)
        self.sensor_type = self._get_sensor_type(sensor_type)
        self.pin = pin
        self.retries = retries
        self.retry_seconds = retry_seconds

    @staticmethod
    def _get_sensor_type(sensor_type: str) -> int:
        import Adafruit_DHT
        sensor_type = sensor_type.upper()
        assert hasattr(Adafruit_DHT, sensor_type), \
            'Unknown sensor type: {}. Supported types: DHT11, DHT22, AM2302'.format(sensor_type)

        return getattr(Adafruit_DHT, sensor_type)

    @action
    def read(self, sensor_type: Optional[str] = None, pin: Optional[int] = None,
             retries: Optional[int] = None, retry_seconds: Optional[int] = None, **__) -> Dict[str, float]:
        """
        Read data from the sensor.

        :param sensor_type: Default ``sensor_type`` override.
        :param pin: Default ``pin`` override.
        :param retries: Default ``retries`` override.
        :param retry_seconds: Default ``retry_seconds`` override.
        :return: A mapping with the measured temperature and humidity. Example:

            .. code-block:: json

                {
                    "humidity": 30.0,
                    "temperature": 25.5
                }

        """
        import Adafruit_DHT
        sensor_type = self._get_sensor_type(sensor_type) if sensor_type else self.sensor_type
        pin = pin or self.pin
        retries = retries or self.retries
        retry_seconds = retry_seconds or self.retry_seconds
        humidity, temperature = Adafruit_DHT.read_retry(sensor=sensor_type,
            pin=pin, retries=retries, delay_seconds=retry_seconds)

        return {
            'humidity': humidity,
            'temperature': temperature,
        }

    @action
    def get_measurement(self) -> Dict[str, float]:
        """
        Get data from the sensor.

        :return: A mapping with the measured temperature and humidity. Example:

            .. code-block:: json

                {
                    "humidity": 30.0,
                    "temperature": 25.5
                }

        """
        return self.read()



# vim:sw=4:ts=4:et:



import time
from typing import Optional, List

from platypush.plugins import Plugin, action


class LastfmPlugin(Plugin):
    """
    Plugin to interact with your Last.FM (https://last.fm) account, update your
    current track and your scrobbles.

    Requires:

        * **pylast** (``pip install pylast``)
    """

    def __init__(self, api_key, api_secret, username, password):
        """
        :param api_key: Last.FM API key, see https://www.last.fm/api
        :type api_key: str

        :param api_secret: Last.FM API secret, see https://www.last.fm/api
        :type api_key: str

        :param username: Last.FM username
        :type api_key: str

        :param password: Last.FM password, used to sign the requests
        :type api_key: str
        """

        import pylast

        super().__init__()

        self.api_key = api_key
        self.api_secret = api_secret
        self.username = username
        self.password = password

        self.lastfm = pylast.LastFMNetwork(
            api_key=self.api_key,
            api_secret=self.api_secret,
            username=self.username,
            password_hash=pylast.md5(self.password),
        )

    @action
    def scrobble(self, artist, title, album=None):
        """
        Scrobble a track to Last.FM

        :param artist: Artist
        :type artist: str
        :param title: Title
        :type title: str
        :param album: Album (optional)
        :type album: str
        """

        self.lastfm.scrobble(
            artist=artist,
            title=title,
            album=album,
            timestamp=int(time.time()),
        )

    @action
    def update_now_playing(self, artist, title, album=None):
        """
        Update the currently playing track

        :param artist: Artist
        :type artist: str
        :param title: Title
        :type title: str
        :param album: Album (optional)
        :type album: str
        """

        self.lastfm.update_now_playing(
            artist=artist,
            title=title,
            album=album,
        )

    @action
    def get_recent_tracks(
        self,
        username: Optional[str] = None,
        limit: int = 10,
        time_from: Optional[int] = None,
        time_to: Optional[int] = None,
    ) -> List[dict]:
        """
        Get a list of recently played tracks.

        :param username: Target username (default: the one registered to this
            plugin).
        :param limit: Maximum number of tracks to be returned (default: 10).
        :param time_from: Return tracks starting from this time
            (as a UNIX timestamp).
        :param time_to: Return tracks starting up to this time
            (as a UNIX timestamp).
        :return: Example:

            .. code-block:: json

                [
                    {
                        "artist": "Led Zeppelin",
                        "title": "Stairway to Heaven",
                        "album": "IV",
                        "timestamp": 1654196137
                    }
                ]

        """

        return [
            {
                'title': track.track.title,
                'album': track.album,
                'timestamp': int(track.timestamp or 0),
                **(
                    {'artist': track.track.artist.name}
                    if track.track.artist
                    else {'artist': None}
                ),
            }
            for track in self.lastfm.get_user(
                username or self.username
            ).get_recent_tracks(
                limit=limit,
                time_from=time_from,
                time_to=time_to,
            )
        ]

    @action
    def get_track(self, artist: str, title: str) -> Optional[dict]:
        """
        Get the information about a track.

        :param artist: Track artist.
        :param title: Track title.
        :return: The retrieved track, with the title corrected if required, if
            it exists on Last.FM. Example:

                .. code-block:: json

                    {
                        "artist": "Led Zeppelin",
                        "title": "Stairway to Heaven",
                        "tags": ["rock", "hard rock", "70s"]
                    }

        """
        track = self.lastfm.get_track(artist, title)
        return {
            'title': track.get_correction(),
            'tags': track.get_tags(),
            **({'artist': track.artist.name} if track.artist else {'artist': None}),
        }

    @action
    def get_similar_tracks(
        self, artist: str, title: str, limit: Optional[int] = None
    ) -> List[dict]:
        """
        Get the tracks that are similar to a specific track.

        :param artist: Track artist.
        :param title: Track title.
        :param limit: Maximum number of suggested tracks to be returned
            (default: None).
        :return: A list of similar tracks, each with a ``match`` score between
            0 and 1. Example:

                .. code-block:: json

                    [
                      {
                        "artist": "Led Zeppelin",
                        "title": "Black Dog",
                        "score": 1.0
                      },
                      {
                        "artist": "Led Zeppelin",
                        "title": "Rock and Roll",
                        "score": 0.89
                      },
                      {
                        "artist": "Eagles",
                        "title": "Hotel California",
                        "score": 0.471
                      },
                      {
                        "artist": "Deep Purple",
                        "title": "Smoke on the Water",
                        "score": 0.393
                      }
                    ]

        """
        track = self.lastfm.get_track(artist, title)
        if not track:
            return []

        return [
            {
                'artist': t.item.artist.name,
                'title': t.item.title,
                'score': t.match,
            }
            for t in track.get_similar(limit=limit)
        ]


# vim:sw=4:ts=4:et:



import os
from typing import Optional, List

# noinspection PyPackageRequirements
from google.cloud import translate_v2 as translate

from platypush.message.response.translate import TranslateResponse
from platypush.plugins import action, Plugin


class GoogleTranslatePlugin(Plugin):
    """
    Plugin to interact with the Google Translate API.
    You'll need a Google Cloud active project and a set of credentials to use this plugin:

        1. Create a project on the `Google Cloud console <https://console.cloud.google.com/projectcreate>`_ if
           you don't have one already.

        2. In the menu navigate to the *Artificial Intelligence* section and select *Translations* and enable the API.

        3. From the menu select *APIs & Services* and create a service account. You can leave role and permissions
           empty.

        4. Create a new private JSON key for the service account and download it. By default platypush will look for the
           credentials file under ``~/.credentials/platypush/google/translate.json``.

    Requires:

        * **google-api-python-client** (``pip install google-api-python-client``)
        * **oauth2client** (``pip install oauth2client``)
        * **google-cloud-translate** (``pip install google-cloud-translate``)

    """

    _maximum_text_length = 2000
    default_credentials_file = os.path.join(os.path.expanduser('~'), '.credentials', 'platypush', 'google',
                                            'translate.json')

    def __init__(self, target_language: str = 'en', credentials_file: Optional[str] = None, **kwargs):
        """
        :param target_language: Default target language (default: 'en').
        :param credentials_file: Google service account JSON credentials file. If none is specified then the plugin will
            search for the credentials file in the following order:

            1. ``~/.credentials/platypush/google/translate.json``
            2. Context from the ``GOOGLE_APPLICATION_CREDENTIALS`` environment variable.
        """
        super().__init__(**kwargs)
        self.target_language = target_language
        self.credentials_file = None

        if credentials_file:
            self.credentials_file = os.path.abspath(os.path.expanduser(credentials_file))
        elif os.path.isfile(self.default_credentials_file):
            self.credentials_file = self.default_credentials_file

        if self.credentials_file:
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = self.credentials_file

    @staticmethod
    def _nearest_delimiter_index(text: str, pos: int) -> int:
        for i in range(min(pos, len(text)-1), -1, -1):
            if text[i] in [' ', '\t', ',', '.', ')', '>']:
                return i
            elif text[i] in ['(', '<']:
                return i-1 if i > 0 else 0

        return 0

    @classmethod
    def _split_text(cls, text: str, length: int = _maximum_text_length) -> List[str]:
        parts = []

        while text:
            i = cls._nearest_delimiter_index(text, length)
            if i == 0:
                parts.append(text)
                text = ''
            else:
                part = text[:i+1]
                if part:
                    parts.append(part.strip())
                text = text[i+1:]

        return parts

    # noinspection PyShadowingBuiltins
    @action
    def translate(self, text: str, target_language: Optional[str] = None, source_language: Optional[str] = None,
                  format: Optional[str] = None) -> TranslateResponse:
        """
        Translate a piece of text or HTML.

        :param text: Input text.
        :param target_language: target_language override.
        :param source_language: source_language (default: auto-detect).
        :param format: Input format (available formats: ``text``, ``html``).
        :return: :class:`platypush.message.response.translate.TranslateResponse`.
        """
        target_language = target_language or self.target_language
        args = {}
        if target_language:
            args['target_language'] = target_language

        client = translate.Client(**args)
        if source_language:
            args['source_language'] = source_language

        inputs = self._split_text(text)
        result = {}

        for input in inputs:
            response = client.translate(input, format_=format, **args)
            if not result:
                result = response
            else:
                # noinspection PyTypeChecker
                result['translatedText'] += ' ' + response['translatedText']

        return TranslateResponse(
            translated_text=result.get('translatedText'),
            source_text=text,
            detected_source_language=result.get('detectedSourceLanguage'),
        )


# vim:sw=4:ts=4:et:



from typing import Optional, Union


class PlatypushException(Exception):
    """
    Base class for all Platypush exceptions.
    """
    def __init__(self, error: Optional[Union[str, Exception]] = None, *args):
        super().__init__(*args)
        self._inner_exception = None
        self._msg = None

        if isinstance(error, str):
            self._msg = error
        elif isinstance(error, Exception):
            self._inner_exception = error
            self._msg = str(error)

    def __str__(self):
        return self._msg



from typing import Any, Dict, Optional, List, Union

from platypush.backend.zwave import ZwaveBackend
from platypush.context import get_backend
from platypush.plugins import action
from platypush.plugins.switch import SwitchPlugin
from platypush.plugins.zwave._base import ZwaveBasePlugin


class ZwavePlugin(ZwaveBasePlugin, SwitchPlugin):
    """
    This plugin interacts with the devices on a Z-Wave network started through the
    :class:`platypush.backend.zwave.ZwaveBackend` backend.

    Requires:

        * **python-openzwave** (``pip install python-openzwave``)
        * The :class:`platypush.backend.zwave.ZwaveBackend` backend configured and running.

    """

    @staticmethod
    def _get_backend() -> ZwaveBackend:
        backend = get_backend('zwave')
        if not backend:
            raise AssertionError('Z-Wave backend not configured')

        return backend

    @classmethod
    def _get_network(cls):
        backend = cls._get_backend()
        if not backend.network:
            backend.start_network()

        return backend.network

    @classmethod
    def _get_controller(cls):
        return cls._get_network().controller

    @action
    def start_network(self):
        backend = self._get_backend()
        backend.start_network()

    @action
    def stop_network(self):
        backend = self._get_backend()
        backend.stop_network()

    @action
    def status(self) -> Dict[str, Any]:
        """
        Get the status of the controller.

        :return: dict
        """
        backend = self._get_backend()
        network = self._get_network()
        controller = self._get_controller()

        return {
            'device': backend.device,
            'state': network.state_str,
            'stats': controller.stats,
        }

    @action
    def add_node(self, do_security=False):
        """
        Start the inclusion process to add a node to the network.

        :param do_security: Whether to initialize the Network Key on the device if it supports the Security CC
        """
        controller = self._get_controller()
        controller.add_node(do_security)

    @action
    def remove_node(self):
        """
        Remove a node from the network.
        """
        controller = self._get_controller()
        controller.remove_node()
        self.write_config()

    @action
    def remove_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Remove a failed node from the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.remove_failed_node(node.node_id)
        self.write_config()

    @action
    def replace_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Replace a failed node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.replace_failed_node(node.node_id)
        self.write_config()

    @action
    def replication_send(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Send node information from the primary to the secondary controller.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.replication_send(node.node_id)

    @action
    def request_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Request a network update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.request_network_update(node.node_id)

    @action
    def request_node_neighbour_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Request a neighbours list update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        controller = self._get_controller()
        node = self._get_node(node_id=node_id, node_name=node_name)
        controller.request_node_neighbor_update(node.node_id)

    @staticmethod
    def value_to_dict(value) -> Dict[str, Any]:
        if not value:
            return {}

        return {
            'command_class': value.node.get_command_class_as_string(value.command_class)
            if value.command_class else None,
            'data': value.data,
            'data_as_string': value.data_as_string,
            'data_items': list(value.data_items) if isinstance(value.data_items, set) else value.data_items,
            'genre': value.genre,
            'help': value.help,
            'home_id': value.home_id,
            'id_on_network': value.id_on_network if value.parent_id is not None and value.command_class is not None
            and value.instance is not None and value.index is not None else None,
            'index': value.index,
            'instance': value.instance,
            'is_polled': value.is_polled,
            'is_read_only': value.is_read_only,
            'is_set': value.is_set,
            'is_write_only': value.is_write_only,
            'label': value.label,
            'last_update': value.last_update,
            'min': value.min,
            'max': value.max,
            'object_id': value.object_id,
            'outdated': value.outdated,
            'parent_id': value.parent_id,
            'poll_intensity': value.poll_intensity,
            'precision': value.precision,
            'type': value.type,
            'units': value.units,
            'use_cache': value.use_cache,
            'value_id': value.value_id,
        }

    @staticmethod
    def group_to_dict(group) -> Dict[str, Any]:
        if not group:
            return {}

        return {
            'index': group.index,
            'label': group.label,
            'max_associations': group.max_associations,
            'associations': group.associations,
        }

    @classmethod
    def node_to_dict(cls, node) -> Dict[str, Any]:
        if not node:
            return {}

        # noinspection PyProtectedMember
        return {
            'node_id': node.node_id,
            'home_id': node.home_id,
            'capabilities': list(node.capabilities),
            'command_classes': [node.get_command_class_as_string(cc) for cc in node.command_classes if cc]
            if hasattr(node, 'command_classes') else [],
            'device_type': node.device_type if hasattr(node, 'device_type') else '',
            'groups': {
                group_id: cls.group_to_dict(group)
                for group_id, group in node.groups.items()
            },
            'is_awake': node.is_awake if hasattr(node, 'is_awake') else False,
            'is_failed': node.is_failed if hasattr(node, 'is_failed') else False,
            'is_beaming_device': node.is_beaming_device if hasattr(node, 'is_beaming_device') else False,
            'is_frequent_listening_device': node.is_frequent_listening_device
            if hasattr(node, 'is_frequent_listening_device') else False,
            'is_info_received': node.is_info_received if hasattr(node, 'is_info_received') else False,
            'is_listening_device': node.is_listening_device if hasattr(node, 'is_listening_device') else False,
            'is_locked': node.is_locked if hasattr(node, 'is_locked') else False,
            'is_ready': node.is_ready if hasattr(node, 'is_ready') else False,
            'is_routing_device': node.is_routing_device if hasattr(node, 'is_routing_device') else False,
            'is_security_device': node.is_security_device if hasattr(node, 'is_security_device') else False,
            'is_sleeping': node.is_sleeping if hasattr(node, 'is_sleeping') else False,
            'last_update': node.last_update if hasattr(node, 'last_update') else None,
            'location': node.location if hasattr(node, 'location') else None,
            'manufacturer_id': node.manufacturer_id if hasattr(node, 'manufacturer_id') else None,
            'manufacturer_name': node.manufacturer_name if hasattr(node, 'manufacturer_name') else None,
            'max_baud_rate': node.max_baud_rate if hasattr(node, 'max_baud_rate') else None,
            'neighbours': [node_id for node_id in node.neighbors] if hasattr(node, 'neighbours') else [],
            'name': node.name if hasattr(node, 'name') else None,
            'outdated': node.outdated if hasattr(node, 'outdated') else False,
            'product_id': node.product_id if hasattr(node, 'product_id') else None,
            'product_name': node.product_name if hasattr(node, 'product_name') else None,
            'product_type': node.product_type if hasattr(node, 'product_type') else None,
            'query_stage': node.query_stage if hasattr(node, 'query_stage') else None,
            'role': node.role if hasattr(node, 'role') else None,
            'security': node.security if hasattr(node, 'security') else None,
            'specific': node.specific if hasattr(node, 'specific') else None,
            'type': node.type if hasattr(node, 'type') else None,
            'use_cache': node.use_cache if hasattr(node, 'use_cache') else False,
            'version': node.version if hasattr(node, 'version') else None,
            'values': {
                value.id_on_network: cls.value_to_dict(value)
                for value_id, value in (node.values or {}).items()
                if value.index is not None
                and value.instance is not None
                and value.command_class
                and value.parent_id is not None
                and value._network
                and value._network.home_id is not None
            } if hasattr(node, 'values') else {},
        }

    def _get_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        assert node_id is not None or node_name is not None, 'Specify either node_id or name'
        nodes = self._get_backend().network.nodes

        if node_id is not None:
            assert node_id in nodes, 'No such node_id: {}'.format(node_id)
            return nodes[node_id]

        nodes = [n for n in nodes.values() if n.name == node_name]
        assert nodes, 'No such node name: {}'.format(node_name)
        return nodes[0]

    def _get_groups(self) -> dict:
        return {
            group_index: group
            for node in self._get_backend().network.nodes.values()
            for group_index, group in node.groups.items()
        }

    def _get_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None):
        assert group_index is not None or group_label is not None, 'Specify either group_index or label'
        groups = self._get_groups()

        if group_index is not None:
            assert group_index in groups, 'No such group_index: {}'.format(group_index)
            return groups[group_index]

        groups = [g for g in groups.values() if g.label == group_label]
        assert groups, 'No such group label: {}'.format(group_label)
        return groups[0]

    def _get_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        assert scene_id is not None or scene_label is not None, 'Specify either scene_id or label'
        scenes = self._get_backend().network.get_scenes()

        if scene_id is not None:
            assert scene_id in scenes, 'No such scene_id: {}'.format(scene_id)
            return scenes[scene_id]

        scenes = [s for s in scenes.values() if s['label'] == scene_label]
        assert scenes, 'No such scene label: {}'.format(scene_label)
        return scenes[0]

    @action
    def get_nodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[str, Any]:
        """
        Get the nodes associated to the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        if node_id is not None or node_name is not None:
            return self.node_to_dict(self._get_node(node_id=node_id, node_name=node_name))

        return {
            node_id: self.node_to_dict(node)
            for node_id, node in self._get_backend().network.nodes.items()
        }

    @action
    def get_node_stats(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[str, Any]:
        """
        Get the statistics of a node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        return node.stats

    @action
    def set_node_name(self, new_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Rename a node on the network.

        :param new_name: New name for the node.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.name = new_name
        self.write_config()

    @action
    def set_node_product_name(self, product_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Set the product name of a node.

        :param product_name: Product name.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.product_name = product_name
        self.write_config()

    @action
    def set_node_manufacturer_name(self, manufacturer_name: str, node_id: Optional[int] = None,
                                   node_name: Optional[str] = None):
        """
        Set the manufacturer name of a node.

        :param manufacturer_name: Manufacturer name.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.manufacturer_name = manufacturer_name
        self.write_config()

    @action
    def set_node_location(self, location: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Set the location of a node.

        :param location: Node location.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.location = location
        self.write_config()

    @action
    def cancel_command(self):
        """
        Cancel the current running command.
        """
        self._get_controller().cancel_command()

    @action
    def kill_command(self):
        """
        Immediately terminate any running command on the controller and release the lock.
        """
        self._get_controller().kill_command()

    @action
    def set_controller_name(self, name: str):
        """
        Set the name of the controller on the network.

        :param name: New controller name.
        """
        self._get_controller().name = name
        self.write_config()

    @action
    def get_capabilities(self) -> List[str]:
        """
        Get the capabilities of the controller.
        """
        return list(self._get_controller().capabilities)

    @action
    def receive_configuration(self):
        """
        Receive the configuration from the primary controller on the network. Requires a primary controller active.
        """
        self._get_controller().receive_configuration()

    @action
    def transfer_primary_role(self):
        """
        Add a new controller to the network and make it the primary.
        The existing primary will become a secondary controller.
        """
        self._get_controller().transfer_primary_role()

    @action
    def heal(self, refresh_routes: bool = False):
        """
        Heal network by requesting nodes rediscover their neighbors.

        :param refresh_routes: Whether to perform return routes initialization (default: ``False``).
        """
        self._get_network().heal(refresh_routes)

    @action
    def switch_all(self, state: bool):
        """
        Switch all the connected devices on/off.

        :param state: True (switch on) or False (switch off).
        """
        self._get_network().switch_all(state)

    @action
    def test(self, count: int = 1):
        """
        Send a number of test messages to every node and record results.

        :param count: The number of test messages to send.
        """
        self._get_network().test(count)

    def _get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                   node_id: Optional[int] = None, node_name: Optional[str] = None, value_label: Optional[str] = None):
        assert (value_id is not None or id_on_network is not None) or \
               ((node_id is not None or node_name is not None) and value_label is not None), \
               'Specify either value_id, id_on_network, or [node_id/node_name, value_label]'

        if value_id is not None:
            return self._get_network().get_value(value_id)
        if id_on_network is not None:
            values = [value
                      for node in self._get_network().nodes.values()
                      for value in node.values.values()
                      if value.id_on_network == id_on_network]

            assert values, 'No such value ID: {}'.format(id_on_network)
            return values[0]

        node = self._get_node(node_id=node_id, node_name=node_name)
        values = [v for v in node.values.values() if v.label == value_label]
        assert values, 'No such value on node "{}": "{}"'.format(node.name, value_label)
        return values[0]

    @action
    def get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None) \
            -> Dict[str, Any]:
        """
        Get a value on the network.

        :param value_id: Select by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        return self._get_value(value_id=value_id, id_on_network=id_on_network,
                               node_id=node_id, node_name=node_name, value_label=value_label).to_dict()

    @action
    def set_value(self, data, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Set a value.

        :param data: Data to set for the value.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        from openzwave.node import ZWaveNode
        value = self._get_value(value_id=value_id, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, value_label=value_label)
        new_val = value.check_data(data)
        assert new_val is not None, 'Invalid value passed to the property'
        node: ZWaveNode = self._get_network().nodes[value.node.node_id]
        node.values[value.value_id].data = new_val
        self.write_config()

    @action
    def set_value_label(self, new_label: str, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                        value_label: Optional[str] = None, node_id: Optional[int] = None,
                        node_name: Optional[str] = None):
        """
        Change the label/name of a value.

        :param new_label: New value label.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, value_label=value_label)
        value.label = new_label
        self.write_config()

    @action
    def node_add_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                       value_label: Optional[str] = None, node_id: Optional[int] = None,
                       node_name: Optional[str] = None):
        """
        Add a value to a node.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node.node_id,
                                value_label=value_label)
        node.add_value(value.value_id)
        self.write_config()

    @action
    def node_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                          value_label: Optional[str] = None, node_id: Optional[int] = None,
                          node_name: Optional[str] = None):
        """
        Remove a value from a node.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node.node_id,
                                value_label=value_label)
        node.remove_value(value.value_id)
        self.write_config()

    @action
    def node_heal(self, node_id: Optional[int] = None, node_name: Optional[str] = None, refresh_routes: bool = False):
        """
        Heal network node by requesting the node to rediscover their neighbours.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param refresh_routes: Whether to perform return routes initialization. (default: ``False``).
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.heal(refresh_routes)

    @action
    def node_update_neighbours(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Ask a node to update its neighbours table.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.neighbor_update()

    @action
    def node_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Update the controller with network information.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.network_update()

    @action
    def node_refresh_info(self, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Fetch up-to-date information about the node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        node.refresh_info()

    def _get_values(self, item: str, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        nodes = [self._get_node(node_id=node_id, node_name=node_name)] if node_id or node_name \
            else self._get_network().nodes.values()

        return {
            value.id_on_network: {
                'node_id': node.node_id,
                'node_name': node.name,
                **self.value_to_dict(value)
            }
            for node in nodes
            for value_id, value in getattr(node, 'get_' + item)().items()
        }

    @action
    def get_dimmers(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the dimmers on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        return self._get_values('dimmers', node_id=node_id, node_name=node_name)

    @action
    def get_node_config(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the configuration values of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        return self._get_values('configs', node_id=node_id, node_name=node_name)

    @action
    def get_battery_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the battery levels of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('battery_levels', node_id=node_id, node_name=node_name)

    @action
    def get_power_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the power levels of this node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('power_levels', node_id=node_id, node_name=node_name)

    @action
    def get_bulbs(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the bulbs/LEDs on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('rgbbulbs', node_id=node_id, node_name=node_name)

    @action
    def get_switches(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the switches on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('switches', node_id=node_id, node_name=node_name)

    @action
    def get_sensors(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the sensors on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('sensors', node_id=node_id, node_name=node_name)

    @action
    def get_doorlocks(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the doorlocks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('doorlocks', node_id=node_id, node_name=node_name)

    @action
    def get_usercodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the usercodes on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('usercodes', node_id=node_id, node_name=node_name)

    @action
    def get_thermostats(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the thermostats on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('thermostats', node_id=node_id, node_name=node_name)

    @action
    def get_protections(self, node_id: Optional[int] = None, node_name: Optional[str] = None) -> Dict[int, Any]:
        """
        Get the protection-compatible devices on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        return self._get_values('protections', node_id=node_id, node_name=node_name)

    @action
    def get_groups(self) -> Dict[int, Any]:
        """
        Get the groups on the network.
        """
        return {
            group_index: self.group_to_dict(group)
            for group_index, group in self._get_groups().items()
        }

    @action
    def get_scenes(self) -> Dict[str, Any]:
        """
        Get the scenes configured on the network.
        """
        network = self._get_network()
        if not network.get_scenes():
            return {}

        return network.scenes_to_dict()

    @action
    def create_scene(self, label: str):
        """
        Create a new scene.

        :param label: Scene label.
        """
        self._get_network().create_scene(label)
        self.write_config()

    @action
    def remove_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        """
        Remove a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        self._get_network().remove_scene(scene.scene_id)
        self.write_config()

    @action
    def activate_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        """
        Activate a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        scene.activate()

    @action
    def set_scene_label(self, new_label: str, scene_id: Optional[int] = None, scene_label: Optional[str] = None):
        """
        Rename a scene/set the scene label.

        :param new_label: New label.
        :param scene_id: Select by scene_id.
        :param scene_label: Select by current scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        scene.label = new_label
        self.write_config()

    @action
    def scene_add_value(self, data: Optional[Any] = None,
                        value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                        value_label: Optional[str] = None, scene_id: Optional[int] = None,
                        scene_label: Optional[str] = None, node_id: Optional[int] = None,
                        node_name: Optional[str] = None):
        """
        Add a value to a scene.

        WARNING: This method actually doesn't work, by own admission of the
        `OpenZWave developer <https://github.com/OpenZWave/python-openzwave/blob/master/src-lib/libopenzwave/libopenzwave.pyx#L4730>`_.


        :param data: Data to set for the value (default: current value data).
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node_id, node_name=node_name,
                                value_label=value_label)
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        data = data if data is not None else value.data
        data = value.check_data(data)
        assert data is not None, 'Invalid value passed to the property'

        scene.add_value(value.value_id, data)
        self.write_config()

    @action
    def scene_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                           value_label: Optional[str] = None, scene_id: Optional[int] = None,
                           scene_label: Optional[str] = None, node_id: Optional[int] = None,
                           node_name: Optional[str] = None):
        """
        Remove a value from a scene.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        """
        value = self._get_value(value_id=value_id, id_on_network=id_on_network, node_id=node_id, node_name=node_name,
                                value_label=value_label)
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        scene.remove_value(value.value_id)
        self.write_config()

    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :return: value_id -> value (as a dict) mapping.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        return {v.value_id: v.to_dict() for v in scene.get_values().items()}

    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label)
        return {v.value_id: v.to_dict() for v in scene.get_values().items()}

    @action
    def create_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Create a handheld button on a device. Only intended for bridge firmware controllers.

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        self._get_controller().create_button(node.node_id, button_id)
        self.write_config()

    @action
    def delete_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Delete a button association from a device. Only intended for bridge firmware controllers.

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        self._get_controller().delete_button(node.node_id, button_id)
        self.write_config()

    @action
    def add_node_to_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None,
                          node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Add a node to a group.

        :param group_index: Select group by group index.
        :param group_label: Select group by group label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by node name.
        :return:
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        group = self._get_group(group_index=group_index, group_label=group_label)
        group.add_association(node.node_id)
        self.write_config()

    @action
    def remove_node_from_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None,
                               node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Remove a node from a group.

        :param group_index: Select group by group index.
        :param group_label: Select group by group label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by node name.
        :return:
        """
        node = self._get_node(node_id=node_id, node_name=node_name)
        group = self._get_group(group_index=group_index, group_label=group_label)
        group.remove_association(node.node_id)
        self.write_config()

    @action
    def create_new_primary(self):
        """
        Create a new primary controller on the network when the previous primary fails.
        """
        self._get_controller().create_new_primary()
        self.write_config()

    @action
    def hard_reset(self):
        """
        Perform a hard reset of the controller. It erases its network configuration settings.
        The controller becomes a primary controller ready to add devices to a new network.
        """
        self._get_controller().hard_reset()

    @action
    def soft_reset(self):
        """
        Perform a soft reset of the controller.
        Resets a controller without erasing its network configuration settings.
        """
        self._get_controller().soft_reset()

    @action
    def write_config(self):
        """
        Store the current configuration of the network to the user directory.
        """
        self._get_network().write_config()

    @action
    def on(self, device: str, *args, **kwargs):
        """
        Turn on a switch on a device.

        :param device: ``id_on_network`` of the value to be switched on.
        """
        # noinspection PyUnresolvedReferences
        switches = self.get_switches().output
        assert device in switches, 'No such id_on_network associated to a switch: {}'.format(device)
        return self.set_value(data=True, id_on_network=device)

    @action
    def off(self, device: str, *args, **kwargs):
        """
        Turn off a switch on a device.

        :param device: ``id_on_network`` of the value to be switched off.
        """
        # noinspection PyUnresolvedReferences
        switches = self.get_switches().output
        assert device in switches, 'No such id_on_network associated to a switch: {}'.format(device)
        return self.set_value(data=False, id_on_network=device)

    @action
    def toggle(self, device: str, *args, **kwargs):
        """
        Toggle a switch on a device.

        :param device: ``id_on_network`` of the value to be toggled.
        """
        # noinspection PyUnresolvedReferences
        switches = self.get_switches().output
        assert device in switches, 'No such id_on_network associated to a switch: {}'.format(device)
        data = switches[device]['data']
        return self.set_value(data=not data, id_on_network=device)

    @property
    def switches(self) -> List[dict]:
        # noinspection PyUnresolvedReferences
        devices = self.get_switches().output.values()
        return [
            {
                'name': '{} - {}'.format(dev.get('node_name', '[No Name]'), dev.get('label', '[No Label]')),
                'on': dev.get('data'),
                'id': dev.get('id_on_network'),
            }
            for dev in devices
        ]


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import Optional

from platypush.message.event import Event


class SunEvent(Event):
    """
    Base class for sun related events (sunrise and sunset).
    """
    def __init__(self, latitude: Optional[float] = None, longitude: Optional[float] = None,
                 time: Optional[datetime] = None, *args, **kwargs):
        """
        :param latitude: Latitude for the sun event.
        :param longitude: Longitude for the sun event.
        :param time: Event timestamp.
        """
        super().__init__(*args, latitude=latitude, longitude=longitude, time=time, **kwargs)
        self.latitude = latitude
        self.longitude = longitude
        self.time = time


class SunriseEvent(SunEvent):
    """
    Class for sunrise events.
    """


class SunsetEvent(SunEvent):
    """
    Class for sunset events.
    """


# vim:sw=4:ts=4:et:



import datetime
import glob
import importlib
import inspect
import logging
import os
import pathlib
import pkgutil
import re
import shutil
import socket
import sys
from typing import Optional

import yaml

from platypush.utils import (
    get_hash,
    is_functional_procedure,
    is_functional_hook,
    is_functional_cron,
)

""" Config singleton instance """
_default_config_instance = None


class Config:
    """
    Configuration base class
    Usage:
        - Initialize config from one of the default paths:
            Config.init()
        - Initialize config from a custom path
            Config.init(config_file_path)
        - Get a value
            Config.get('foo')
    """

    """
    Default config file locations:
        - $HOME/.config/platypush/config.yaml
        - /etc/platypush/config.yaml
    """
    _cfgfile_locations = [
        os.path.join(os.path.expanduser('~'), '.config', 'platypush', 'config.yaml'),
        os.path.join(os.sep, 'etc', 'platypush', 'config.yaml'),
    ]

    _default_constants = {
        'today': datetime.date.today,
        'now': datetime.datetime.now,
    }

    _workdir_location = os.path.join(
        os.path.expanduser('~'), '.local', 'share', 'platypush'
    )
    _included_files = set()

    def __init__(self, cfgfile=None):
        """
        Constructor. Always use the class as a singleton (i.e. through
        Config.init), you won't probably need to call the constructor directly
        Params:
            cfgfile -- Config file path (default: retrieve the first
                       available location in _cfgfile_locations)
        """

        if cfgfile is None:
            cfgfile = self._get_default_cfgfile()

        if cfgfile is None:
            cfgfile = self._create_default_config()

        self._cfgfile = os.path.abspath(os.path.expanduser(cfgfile))
        self._config = self._read_config_file(self._cfgfile)

        if 'token' in self._config:
            self._config['token_hash'] = get_hash(self._config['token'])

        if 'workdir' not in self._config:
            self._config['workdir'] = self._workdir_location
        os.makedirs(self._config['workdir'], exist_ok=True)

        if 'scripts_dir' not in self._config:
            self._config['scripts_dir'] = os.path.join(
                os.path.dirname(cfgfile), 'scripts'
            )
        os.makedirs(self._config['scripts_dir'], mode=0o755, exist_ok=True)

        if 'dashboards_dir' not in self._config:
            self._config['dashboards_dir'] = os.path.join(
                os.path.dirname(cfgfile), 'dashboards'
            )
        os.makedirs(self._config['dashboards_dir'], mode=0o755, exist_ok=True)

        init_py = os.path.join(self._config['scripts_dir'], '__init__.py')
        if not os.path.isfile(init_py):
            with open(init_py, 'w') as f:
                f.write('# Auto-generated __init__.py - do not remove\n')

        # Include scripts_dir parent in sys.path so members can be imported in scripts
        # through the `scripts` package
        scripts_parent_dir = str(
            pathlib.Path(self._config['scripts_dir']).absolute().parent
        )
        sys.path = [scripts_parent_dir] + sys.path

        self._config['db'] = self._config.get(
            'main.db',
            {
                'engine': 'sqlite:///'
                + os.path.join(
                    os.path.expanduser('~'), '.local', 'share', 'platypush', 'main.db'
                )
            },
        )

        logging_config = {
            'level': logging.INFO,
            'stream': sys.stdout,
            'format': '%(asctime)-15s|%(levelname)5s|%(name)s|%(message)s',
        }

        if 'logging' in self._config:
            for (k, v) in self._config['logging'].items():
                if k == 'filename':
                    logfile = os.path.expanduser(v)
                    logdir = os.path.dirname(logfile)
                    try:
                        os.makedirs(logdir, exist_ok=True)
                    except Exception as e:
                        print(
                            'Unable to create logs directory {}: {}'.format(
                                logdir, str(e)
                            )
                        )

                    v = logfile
                    del logging_config['stream']
                elif k == 'level':
                    try:
                        v = int(v)
                    except ValueError:
                        v = getattr(logging, v.upper())

                logging_config[k] = v

        self._config['logging'] = logging_config

        if 'device_id' not in self._config:
            self._config['device_id'] = socket.gethostname()

        if 'environment' in self._config:
            for k, v in self._config['environment'].items():
                os.environ[k] = str(v)

        self.backends = {}
        self.plugins = {}
        self.event_hooks = {}
        self.procedures = {}
        self.constants = {}
        self.cronjobs = {}
        self.dashboards = {}
        self._plugin_manifests = {}
        self._backend_manifests = {}

        self._init_manifests()
        self._init_constants()
        self._load_scripts()
        self._init_components()
        self._init_dashboards(self._config['dashboards_dir'])

    def _create_default_config(self):
        cfg_mod_dir = os.path.dirname(os.path.abspath(__file__))
        cfgfile = self._cfgfile_locations[0]
        cfgdir = pathlib.Path(cfgfile).parent
        cfgdir.mkdir(parents=True, exist_ok=True)
        for cfgfile in glob.glob(os.path.join(cfg_mod_dir, 'config*.yaml')):
            shutil.copy(cfgfile, str(cfgdir))

        return cfgfile

    def _read_config_file(self, cfgfile):
        cfgfile_dir = os.path.dirname(os.path.abspath(os.path.expanduser(cfgfile)))

        config = {}

        with open(cfgfile, 'r') as fp:
            file_config = yaml.safe_load(fp)

        if not file_config:
            return config

        for section in file_config:
            if section == 'include':
                include_files = (
                    file_config[section]
                    if isinstance(file_config[section], list)
                    else [file_config[section]]
                )

                for include_file in include_files:
                    if not include_file:
                        continue
                    if not os.path.isabs(include_file):
                        include_file = os.path.join(cfgfile_dir, include_file)
                    self._included_files.add(include_file)

                    included_config = self._read_config_file(include_file)
                    for incl_section in included_config.keys():
                        config[incl_section] = included_config[incl_section]
            elif section == 'scripts_dir':
                assert isinstance(file_config[section], str)
                config['scripts_dir'] = os.path.abspath(
                    os.path.expanduser(file_config[section])
                )
            else:
                section_config = file_config.get(section, {}) or {}
                if not (
                    hasattr(section_config, 'get') and section_config.get('disabled')
                ):
                    config[section] = section_config

        return config

    def _load_module(self, modname: str, prefix: Optional[str] = None):
        try:
            module = importlib.import_module(modname)
        except Exception as e:
            print(
                'Unhandled exception while importing module {}: {}'.format(
                    modname, str(e)
                )
            )
            return

        prefix = modname + '.' if prefix is None else prefix
        self.procedures.update(
            **{
                prefix + name: obj
                for name, obj in inspect.getmembers(module)
                if is_functional_procedure(obj)
            }
        )

        self.event_hooks.update(
            **{
                prefix + name: obj
                for name, obj in inspect.getmembers(module)
                if is_functional_hook(obj)
            }
        )

        self.cronjobs.update(
            **{
                prefix + name: obj
                for name, obj in inspect.getmembers(module)
                if is_functional_cron(obj)
            }
        )

    def _load_scripts(self):
        scripts_dir = self._config['scripts_dir']
        sys_path = sys.path.copy()
        sys.path = [scripts_dir] + sys.path
        scripts_modname = os.path.basename(scripts_dir)
        self._load_module(scripts_modname, prefix='')

        for _, modname, _ in pkgutil.walk_packages(
            path=[scripts_dir], onerror=lambda _: None
        ):
            self._load_module(modname)

        sys.path = sys_path

    def _init_components(self):
        for key in self._config.keys():
            if (
                key.startswith('backend.')
                and '.'.join(key.split('.')[1:]) in self._backend_manifests
            ):
                backend_name = '.'.join(key.split('.')[1:])
                self.backends[backend_name] = self._config[key]
            elif key.startswith('event.hook.'):
                hook_name = '.'.join(key.split('.')[2:])
                self.event_hooks[hook_name] = self._config[key]
            elif key.startswith('cron.'):
                cron_name = '.'.join(key.split('.')[1:])
                self.cronjobs[cron_name] = self._config[key]
            elif key.startswith('procedure.'):
                tokens = key.split('.')
                _async = bool(len(tokens) > 2 and tokens[1] == 'async')
                procedure_name = '.'.join(tokens[2:] if len(tokens) > 2 else tokens[1:])
                args = []
                m = re.match(r'^([^(]+)\(([^)]+)\)\s*', procedure_name)

                if m:
                    procedure_name = m.group(1).strip()
                    args = [
                        arg.strip()
                        for arg in m.group(2).strip().split(',')
                        if arg.strip()
                    ]

                self.procedures[procedure_name] = {
                    '_async': _async,
                    'actions': self._config[key],
                    'args': args,
                }
            elif key in self._plugin_manifests:
                self.plugins[key] = self._config[key]

    def _init_manifests(self, base_dir: Optional[str] = None):
        if not base_dir:
            base_dir = os.path.abspath(os.path.join(__file__, '..', '..'))
            plugins_dir = os.path.join(base_dir, 'plugins')
            backends_dir = os.path.join(base_dir, 'backend')
            self._init_manifests(plugins_dir)
            self._init_manifests(backends_dir)
        else:
            manifests_map = (
                self._plugin_manifests
                if base_dir.endswith('plugins')
                else self._backend_manifests
            )
            for mf in pathlib.Path(base_dir).rglob('manifest.yaml'):
                with open(mf, 'r') as f:
                    manifest = yaml.safe_load(f)['manifest']
                    comp_name = '.'.join(manifest['package'].split('.')[2:])
                    manifests_map[comp_name] = manifest

    def _init_constants(self):
        if 'constants' in self._config:
            self.constants = self._config['constants']

        for (key, value) in self._default_constants.items():
            self.constants[key] = value

    def _get_dashboard(
        self, name: str, dashboards_dir: Optional[str] = None
    ) -> Optional[str]:
        dashboards_dir = dashboards_dir or self._config['dashboards_dir']
        assert dashboards_dir
        abspath = os.path.join(dashboards_dir, name + '.xml')
        if not os.path.isfile(abspath):
            return

        with open(abspath, 'r') as fp:
            return fp.read()

    def _get_dashboards(self, dashboards_dir: Optional[str] = None) -> dict:
        dashboards = {}
        dashboards_dir = dashboards_dir or self._config['dashboards_dir']
        assert dashboards_dir

        for f in os.listdir(dashboards_dir):
            abspath = os.path.join(dashboards_dir, f)
            if not os.path.isfile(abspath) or not abspath.endswith('.xml'):
                continue

            name = f.split('.xml')[0]
            dashboards[name] = self._get_dashboard(name, dashboards_dir)

        return dashboards

    @staticmethod
    def get_dashboard(name: str, dashboards_dir: Optional[str] = None) -> Optional[str]:
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance._get_dashboard(name, dashboards_dir)

    @classmethod
    def get_dashboards(cls, dashboards_dir: Optional[str] = None) -> dict:
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance._get_dashboards(dashboards_dir)

    def _init_dashboards(self, dashboards_dir: str):
        self.dashboards = self._get_dashboards(dashboards_dir)

    @staticmethod
    def get_backends():
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance.backends

    @staticmethod
    def get_plugins():
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance.plugins

    @staticmethod
    def get_event_hooks():
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance.event_hooks

    @staticmethod
    def get_procedures():
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance.procedures

    @staticmethod
    def get_constants():
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        constants = {}

        for name in _default_config_instance.constants.keys():
            constants[name] = Config.get_constant(name)
        return constants

    @staticmethod
    def get_constant(name):
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()

        if name not in _default_config_instance.constants:
            return None
        value = _default_config_instance.constants[name]
        return value() if callable(value) else value

    @staticmethod
    def get_cronjobs():
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()
        return _default_config_instance.cronjobs

    @classmethod
    def _get_default_cfgfile(cls):
        for location in cls._cfgfile_locations:
            if os.path.isfile(location):
                return location

    @staticmethod
    def init(cfgfile=None):
        """
        Initializes the config object singleton
        Params:
            cfgfile -- path to the config file - default: _cfgfile_locations
        """
        global _default_config_instance
        _default_config_instance = Config(cfgfile)

    @staticmethod
    def get(key: Optional[str] = None):
        """
        Get a config value or the whole configuration object.

        :param key: Configuration entry to get (default: all entries).
        """
        global _default_config_instance
        if _default_config_instance is None:
            _default_config_instance = Config()

        if key:
            return _default_config_instance._config.get(key)

        return _default_config_instance._config


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class NextCloudActivityEvent(Event):
    def __init__(self, activity_id: int, activity_type: str, *args, **kwargs):
        super().__init__(*args, activity_id=activity_id, activity_type=activity_type, **kwargs)


# vim:sw=4:ts=4:et:



import json
import os

from platypush.utils import get_redis
from . import status_queue

environ_variables = [
    'PLAYER_EVENT',
    'TRACK_ID',
    'OLD_TRACK_ID',
    'DURATION_MS',
    'POSITION_MS',
    'VOLUME',
]


def on_librespot_event():
    get_redis().rpush(status_queue, json.dumps({
        var: os.environ[var]
        for var in environ_variables
        if var in os.environ
    }))


if __name__ == '__main__':
    on_librespot_event()



import json
import os
import subprocess
import threading
from typing import Optional, Dict, Any

from platypush.backend import Backend
from platypush.common.spotify import SpotifyMixin
from platypush.config import Config
from platypush.message.event.music import MusicPlayEvent, MusicPauseEvent, MusicStopEvent, \
    NewPlayingTrackEvent, SeekChangeEvent, VolumeChangeEvent
from platypush.utils import get_redis

from .event import status_queue


class MusicSpotifyBackend(Backend, SpotifyMixin):
    """
    This backend uses `librespot <https://github.com/librespot-org/librespot>`_ to turn Platypush into an audio client
    compatible with Spotify Connect and discoverable by a device running a Spotify client or app. It can be used to
    stream Spotify through the Platypush host. After the backend has started, you should see a new entry in the
    Spotify Connect devices list in your app.

    Triggers:

        * :class:`platypush.message.event.music.MusicPlayEvent` if the playback state changed to play
        * :class:`platypush.message.event.music.MusicPauseEvent` if the playback state changed to pause
        * :class:`platypush.message.event.music.MusicStopEvent` if the playback state changed to stop
        * :class:`platypush.message.event.music.NewPlayingTrackEvent` if a new track is being played
        * :class:`platypush.message.event.music.VolumeChangeEvent` if the volume changes

    Requires:

        * **librespot**. Consult the `README <https://github.com/librespot-org/librespot>`_ for instructions.

    """

    def __init__(self,
                 librespot_path: str = 'librespot',
                 device_name: Optional[str] = None,
                 device_type: str = 'speaker',
                 audio_backend: str = 'alsa',
                 audio_device: Optional[str] = None,
                 mixer: str = 'softvol',
                 mixer_name: str = 'PCM',
                 mixer_card: str = 'default',
                 mixer_index: int = 0,
                 volume: int = 100,
                 volume_ctrl: str = 'linear',
                 bitrate: int = 160,
                 autoplay: bool = False,
                 disable_gapless: bool = False,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 client_id: Optional[str] = None,
                 client_secret: Optional[str] = None,
                 proxy: Optional[str] = None,
                 ap_port: Optional[int] = None,
                 disable_discovery: bool = False,
                 cache_dir: Optional[str] = None,
                 system_cache_dir: Optional[str] = None,
                 disable_audio_cache=False,
                 enable_volume_normalization: bool = False,
                 normalization_method: str = 'dynamic',
                 normalization_pre_gain: Optional[float] = None,
                 normalization_threshold: float = -1.,
                 normalization_attack: int = 5,
                 normalization_release: int = 100,
                 normalization_knee: float = 1.,
                 **kwargs):
        """
        :param librespot_path: Librespot path/executable name (default: ``librespot``).
        :param device_name: Device name (default: same as configured Platypush ``device_id`` or hostname).
        :param device_type: Device type to be shown in the icon. Available types:
            ``unknown``, ``computer``, ``tablet``, ``smartphone``, ``speaker``, ``tv``,
            ``avr`` (Audio/Video Receiver), ``stb`` (Set-Top Box), or ``audiodongle`` (default: ``speaker``).
        :param audio_backend: Audio backend to be used. Supported values:
            ``alsa``, ``portaudio``, ``pulseaudio``, ``jackaudio``, ``gstreamer``, ``rodio``, ``rodiojack``,
            ``sdl`` (default: ``alsa``).
        :param audio_device: Output audio device. Type ``librespot --device ?`` to get a list of the available devices.
        :param mixer: Mixer to be used to control the volume. Supported values: ``alsa`` or ``softvol``
            (default: ``softvol``).
        :param mixer_name: Mixer name if using the ALSA mixer. Supported values: ``PCM`` or ``Master``
            (default: ``PCM``).
        :param mixer_card: ALSA mixer output card, as reported by ``aplay -l`` (default: ``default``).
        :param mixer_index: ALSA card index, as reported by ``aplay -l`` (default: 0).
        :param volume: Initial volume, as an integer between 0 and 100 if ``volume_ctrl=linear`` or in dB if
            ``volume_ctrl=logarithmic``.
        :param volume_ctrl: Volume control scale. Supported values: ``linear`` and ``logarithmic``
            (default: ``linear``).
        :param bitrate: Audio bitrate. Choose 320 for maximum quality (default: 160).
        :param autoplay: Play similar tracks when the queue ends (default: False).
        :param disable_gapless: Disable gapless audio (default: False).
        :param username: Spotify user/device username (used if you want to enable Spotify Connect remotely).
        :param password: Spotify user/device password (used if you want to enable Spotify Connect remotely).
        :param client_id: Spotify client ID, required if you want to retrieve track and album info through the
            Spotify Web API. You can generate one by creating a Spotify app `here <https://developer.spotify.com/>`.
        :param client_secret: Spotify client secret, required if you want to retrieve track and album info
            through the Spotify Web API. You can generate one by creating a Spotify app
            `here <https://developer.spotify.com/>`.
        :param proxy: Optional HTTP proxy configuration.
        :param ap_port: Spotify AP port to be used (default: default ports, i.e. 80, 443 and 4070).
        :param disable_discovery: Disable discovery mode.
        :param cache_dir: Data files cache directory.
        :param system_cache_dir: System cache directory - it includes audio settings and credentials.
        :param disable_audio_cache: Disable audio caching (default: False).
        :param enable_volume_normalization: Play all the tracks at about the same volume (default: False).
        :param normalization_method: If ``enable_volume_normalization=True``, this setting specifies the volume
            normalization method. Supported values: ``basic``, ``dynamic`` (default: ``dynamic``).
        :param normalization_pre_gain: Pre-gain applied to volume normalization if ``enable_volume_normalization=True``,
            expressed in dB.
        :param normalization_threshold: If ``enable_volume_normalization=True``, this setting specifies the
            normalization threshold (in dBFS) to prevent audio clipping (default: -1.0).
        :param normalization_attack: If ``enable_volume_normalization=True``, this setting specifies the attack time
            (in ms) during which the dynamic limiter is reducing the gain (default: 5).
        :param normalization_release: If ``enable_volume_normalization=True``, this setting specifies the release time
            (in ms) for the dynamic limiter to restore the gain (default: 100).
        :param normalization_knee: Knee steepness of the dynamic limiter (default: 1.0).
        """
        Backend.__init__(self, **kwargs)
        SpotifyMixin.__init__(self, client_id=client_id, client_secret=client_secret)
        self.device_name = device_name or Config.get('device_id')
        self._librespot_args = [
            librespot_path, '--name', self.device_name, '--backend', audio_backend,
            '--device-type', device_type, '--mixer', mixer, '--alsa-mixer-control', mixer_name,
            '--initial-volume', str(volume), '--volume-ctrl', volume_ctrl, '--bitrate', str(bitrate),
            '--emit-sink-events', '--onevent', 'python -m platypush.backend.music.spotify.event',
        ]

        if audio_device:
            self._librespot_args += ['--alsa-mixer-device', audio_device]
        else:
            self._librespot_args += [
                '--alsa-mixer-device', mixer_card, '--alsa-mixer-index', str(mixer_index)
            ]
        if autoplay:
            self._librespot_args += ['--autoplay']
        if disable_gapless:
            self._librespot_args += ['--disable-gapless']
        if disable_discovery:
            self._librespot_args += ['--disable-discovery']
        if disable_audio_cache:
            self._librespot_args += ['--disable-audio-cache']
        if proxy:
            self._librespot_args += ['--proxy', proxy]
        if ap_port:
            self._librespot_args += ['--ap-port', str(ap_port)]
        if cache_dir:
            self._librespot_args += ['--cache', os.path.expanduser(cache_dir)]
        if system_cache_dir:
            self._librespot_args += ['--system-cache', os.path.expanduser(system_cache_dir)]
        if enable_volume_normalization:
            self._librespot_args += [
                '--enable-volume-normalisation', '--normalisation-method', normalization_method,
                '--normalisation-threshold', str(normalization_threshold), '--normalisation-attack',
                str(normalization_attack), '--normalisation-release', str(normalization_release),
                '--normalisation-knee', str(normalization_knee),
            ]

            if normalization_pre_gain:
                self._librespot_args += ['--normalisation-pregain', str(normalization_pre_gain)]

        self._librespot_dump_args = self._librespot_args.copy()
        if username and password:
            self._librespot_args += ['--username', username, '--password', password]
            self._librespot_dump_args += ['--username', username, '--password', '*****']

        self._librespot_proc: Optional[subprocess.Popen] = None
        self._status_thread: Optional[threading.Thread] = None

        self.status: Dict[str, Any] = {
            'state': 'stop',
            'volume': None,
            'time': None,
            'elapsed': None,
        }

        self.track = {
            'file': None,
            'url': None,
            'uri': None,
            'time': None,
            'artist': None,
            'album': None,
            'title': None,
            'date': None,
            'track': None,
            'id': None,
            'x-albumuri': None,
        }

    def run(self):
        super().run()
        self._status_thread = threading.Thread(target=self._get_status_check_loop())
        self._status_thread.start()

        while not self.should_stop():
            self.logger.info(
                f'Starting music.spotify backend. Librespot command line: {self._librespot_dump_args}'
            )

            try:
                self._librespot_proc = subprocess.Popen(self._librespot_args)

                while not self.should_stop():
                    try:
                        if self._librespot_proc:
                            self._librespot_proc.wait(timeout=1.0)
                    except subprocess.TimeoutExpired:
                        pass
            except Exception as e:
                self.logger.exception(e)
                continue

    def _get_status_check_loop(self):
        def loop():
            redis = get_redis()

            while not self.should_stop():
                msg = redis.blpop(status_queue, timeout=1)
                if not msg:
                    continue

                self._process_status_msg(json.loads(msg[1]))

        return loop

    def _process_status_msg(self, status):
        event_type = status.get('PLAYER_EVENT')
        volume = int(status['VOLUME'])/655.35 if status.get('VOLUME') is not None else None
        track_id = status.get('TRACK_ID')
        old_track_id = status.get('OLD_TRACK_ID', self.track['id'])
        duration = int(status['DURATION_MS'])/1000. if status.get('DURATION_MS') is not None else None
        elapsed = int(status['POSITION_MS'])/1000. if status.get('POSITION_MS') is not None else None

        if volume is not None:
            self.status['volume'] = volume
        if duration is not None:
            self.status['time'] = duration
        if elapsed is not None:
            self.status['elapsed'] = elapsed
        if track_id and track_id != old_track_id:
            self.track = self.spotify_get_track(track_id)

        if event_type == 'playing':
            self.status['state'] = 'play'
        elif event_type == 'paused':
            self.status['state'] = 'pause'
        elif event_type in ['stopped', 'started']:
            self.status['state'] = 'stop'

        event_args = {
            'status': self.status,
            'track': self.track,
            'plugin_name': 'music.spotify',
            'player': self.device_name,
        }

        if event_type == 'volume_set':
            self.bus.post(VolumeChangeEvent(volume=volume, **event_args))
        if elapsed is not None:
            self.bus.post(SeekChangeEvent(position=elapsed, **event_args))
        if track_id and track_id != old_track_id:
            self.bus.post(NewPlayingTrackEvent(**event_args))
        if event_type == 'playing':
            self.bus.post(MusicPlayEvent(**event_args))
        elif event_type == 'paused':
            self.bus.post(MusicPauseEvent(**event_args))
        elif event_type == 'stopped':
            self.bus.post(MusicStopEvent(**event_args))

    def on_stop(self):
        if self._librespot_proc:
            self.logger.info('Terminating librespot')
            self._librespot_proc.terminate()

            try:
                self._librespot_proc.wait(timeout=5.)
            except subprocess.TimeoutExpired:
                self.logger.warning('Librespot has not yet terminated: killing it')
                self._librespot_proc.kill()

            self._librespot_proc = None

        if self._status_thread.is_alive():
            self.logger.info('Waiting for the status check thread to terminate')
            self._status_thread.join(timeout=10)



from platypush.backend.sensor import SensorBackend


class SensorDhtBackend(SensorBackend):
    """
    Backend to poll a DHT11/DHT22/AM2302 temperature/humidity sensor.

    Requires:

        * ``Adafruit_Python_DHT`` (``pip install git+https://github.com/adafruit/Adafruit_Python_DHT.git``)
        * The ``gpio.sensor.dht`` plugin configured and enabled.

    Triggers:

        * :class:`platypush.message.event.sensor.SensorDataChangeEvent` if the measurements of a sensor have changed
        * :class:`platypush.message.event.sensor.SensorDataAboveThresholdEvent` if the measurements of a sensor have
            gone above a configured threshold
        * :class:`platypush.message.event.sensor.SensorDataBelowThresholdEvent` if the measurements of a sensor have
            gone below a configured threshold

    """

    def __init__(self, temperature: bool = True, humidity: bool = True, **kwargs):
        """
        :param temperature: Enable temperature sensor poll.
        :param humidity: Enable humidity sensor poll.
        """

        enabled_sensors = {
            'humidity': humidity,
            'temperature': temperature,
        }

        super().__init__(plugin='gpio.sensor.dht', enabled_sensors=enabled_sensors, **kwargs)


# vim:sw=4:ts=4:et:



import datetime
import os
import threading

from typing import Optional, List

import requests
from sqlalchemy import create_engine, Column, String, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session

from platypush.backend import Backend
from platypush.config import Config
from platypush.message.event.github import GithubPushEvent, GithubCommitCommentEvent, GithubCreateEvent, \
    GithubDeleteEvent, GithubEvent, GithubForkEvent, GithubWikiEvent, GithubIssueCommentEvent, GithubIssueEvent, \
    GithubMemberEvent, GithubPublicEvent, GithubPullRequestEvent, GithubPullRequestReviewCommentEvent, \
    GithubReleaseEvent, GithubSponsorshipEvent, GithubWatchEvent

Base = declarative_base()
Session = scoped_session(sessionmaker())


class GithubResource(Base):
    """
    Models the GithubLastEvent table, containing the timestamp where a certain URL was last checked.
    """

    __tablename__ = 'GithubLastEvent'
    uri = Column(String, primary_key=True)
    last_updated_at = Column(DateTime)


class GithubBackend(Backend):
    """
    This backend monitors for notifications and events either on Github user, organization or repository level.
    You'll need a Github personal access token to use the service. To get one:

        - Access your Github profile settings
        - Select *Developer Settings*
        - Select *Personal access tokens*
        - Click *Generate new token*

    This backend requires the following permissions:

        - ``repo``
        - ``notifications``
        - ``read:org`` if you want to access repositories on organization level.

    Triggers:

        - :class:`platypush.message.event.github.GithubPushEvent` when a new push is created.
        - :class:`platypush.message.event.github.GithubCommitCommentEvent` when a new commit comment is created.
        - :class:`platypush.message.event.github.GithubCreateEvent` when a tag or branch is created.
        - :class:`platypush.message.event.github.GithubDeleteEvent` when a tag or branch is deleted.
        - :class:`platypush.message.event.github.GithubForkEvent` when a user forks a repository.
        - :class:`platypush.message.event.github.GithubWikiEvent` when new activity happens on a repository wiki.
        - :class:`platypush.message.event.github.GithubIssueCommentEvent` when new activity happens on an issue comment.
        - :class:`platypush.message.event.github.GithubIssueEvent` when new repository issue activity happens.
        - :class:`platypush.message.event.github.GithubMemberEvent` when new repository collaborators activity happens.
        - :class:`platypush.message.event.github.GithubPublicEvent` when a repository goes public.
        - :class:`platypush.message.event.github.GithubPullRequestEvent` when new pull request related activity happens.
        - :class:`platypush.message.event.github.GithubPullRequestReviewCommentEvent` when activity happens on a pull
            request commit.
        - :class:`platypush.message.event.github.GithubReleaseEvent` when a new release happens.
        - :class:`platypush.message.event.github.GithubSponsorshipEvent` when new sponsorship related activity happens.
        - :class:`platypush.message.event.github.GithubWatchEvent` when someone stars/starts watching a repository.
        - :class:`platypush.message.event.github.GithubEvent` for any event that doesn't fall in the above categories
            (``event_type`` will be set accordingly).

    """

    _base_url = 'https://api.github.com'

    def __init__(self, user: str, user_token: str, repos: Optional[List[str]] = None, org: Optional[str] = None,
                 poll_seconds: int = 60, max_events_per_scan: Optional[int] = 10, *args, **kwargs):
        """
        If neither ``repos`` nor ``org`` is specified then the backend will monitor all new events on user level.

        :param user: Github username.
        :param user_token: Github personal access token.
        :param repos: List of repos to be monitored - if a list is provided then only these repositories will be
            monitored for events. Repositories should be passed in the format ``username/repository``.
        :param org: Organization to be monitored - if provided then only this organization will be monitored for events.
        :param poll_seconds: How often the backend should check for new events, in seconds (default: 60).
        :param max_events_per_scan: Maximum number of events per resource that will be triggered if there is a large
            number of events/notification since the last check (default: 10). Specify 0 or null for no limit.
        """
        super().__init__(*args, **kwargs)
        self._last_text: Optional[str] = None
        self.user = user
        self.user_token = user_token
        self.repos = repos or []
        self.org = org
        self.poll_seconds = poll_seconds
        self.db_lock = threading.RLock()
        self.workdir = os.path.join(os.path.expanduser(Config.get('workdir')), 'github')
        self.dbfile = os.path.join(self.workdir, 'github.db')
        self.max_events_per_scan = max_events_per_scan

        os.makedirs(os.path.dirname(self.dbfile), exist_ok=True)
        self._init_db()

    def _request(self, uri: str, method: str = 'get') -> dict:
        method = getattr(requests, method.lower())
        return method(self._base_url + uri, auth=(self.user, self.user_token),
                      headers={'Accept': 'application/vnd.github.v3+json'}).json()

    def _init_db(self):
        engine = create_engine('sqlite:///{}'.format(self.dbfile), connect_args={'check_same_thread': False})
        Base.metadata.create_all(engine)
        Session.configure(bind=engine)

    @staticmethod
    def _to_datetime(time_string: str) -> datetime.datetime:
        """ Convert ISO 8061 string format with leading 'Z' into something understandable by Python """
        return datetime.datetime.fromisoformat(time_string[:-1] + '+00:00')

    @staticmethod
    def _get_or_create_resource(uri: str, session: Session) -> GithubResource:
        record = session.query(GithubResource).filter_by(uri=uri).first()
        if record is None:
            record = GithubResource(uri=uri)
            session.add(record)
            session.commit()

        return record

    def _get_last_event_time(self, uri: str):
        with self.db_lock:
            record = self._get_or_create_resource(uri=uri, session=Session())
            return record.last_updated_at.replace(tzinfo=datetime.timezone.utc) if record.last_updated_at else None

    def _update_last_event_time(self, uri: str, last_updated_at: datetime.datetime):
        with self.db_lock:
            session = Session()
            record = self._get_or_create_resource(uri=uri, session=session)
            record.last_updated_at = last_updated_at
            session.add(record)
            session.commit()

    @classmethod
    def _parse_event(cls, event: dict) -> GithubEvent:
        event_mapping = {
            'PushEvent': GithubPushEvent,
            'CommitCommentEvent': GithubCommitCommentEvent,
            'CreateEvent': GithubCreateEvent,
            'DeleteEvent': GithubDeleteEvent,
            'ForkEvent': GithubForkEvent,
            'GollumEvent': GithubWikiEvent,
            'IssueCommentEvent': GithubIssueCommentEvent,
            'IssuesEvent': GithubIssueEvent,
            'MemberEvent': GithubMemberEvent,
            'PublicEvent': GithubPublicEvent,
            'PullRequestEvent': GithubPullRequestEvent,
            'PullRequestReviewCommentEvent': GithubPullRequestReviewCommentEvent,
            'ReleaseEvent': GithubReleaseEvent,
            'SponsorshipEvent': GithubSponsorshipEvent,
            'WatchEvent': GithubWatchEvent,
        }

        event_type = event_mapping[event['type']] if event['type'] in event_mapping else GithubEvent
        return event_type(event_type=event['type'], actor=event['actor'], repo=event.get('repo', {}),
                          payload=event['payload'], created_at=cls._to_datetime(event['created_at']))

    def _events_monitor(self, uri: str, method: str = 'get'):
        def thread():
            while not self.should_stop():
                try:
                    events = self._request(uri, method)
                    if not events:
                        continue

                    last_event_time = self._get_last_event_time(uri)
                    new_last_event_time = last_event_time
                    fired_events = []

                    for event in events:
                        if self.max_events_per_scan and len(fired_events) >= self.max_events_per_scan:
                            break

                        event_time = self._to_datetime(event['created_at'])
                        if last_event_time and event_time <= last_event_time:
                            break
                        if not new_last_event_time or event_time > new_last_event_time:
                            new_last_event_time = event_time

                        fired_events.append(self._parse_event(event))

                    for event in fired_events:
                        self.bus.post(event)

                    self._update_last_event_time(uri=uri, last_updated_at=new_last_event_time)
                except Exception as e:
                    self.logger.warning('Encountered exception while fetching events from {}: {}'.format(
                        uri, str(e)))
                    self.logger.exception(e)
                finally:
                    if self.wait_stop(timeout=self.poll_seconds):
                        break

        return thread

    def run(self):
        self.logger.info('Starting Github backend')
        monitors = []

        if self.repos:
            for repo in self.repos:
                monitors.append(threading.Thread(target=self._events_monitor('/networks/{repo}/events'.format(repo=repo))))
        if self.org:
            monitors.append(threading.Thread(target=self._events_monitor('/orgs/{org}/events'.format(org=self.org))))

        if not (self.repos or self.org):
            monitors.append(threading.Thread(target=self._events_monitor('/users/{user}/events'.format(user=self.user))))

        for monitor in monitors:
            monitor.start()

        self.logger.info('Started Github backend')
        for monitor in monitors:
            monitor.join()

        self.logger.info('Github backend terminated')

# vim:sw=4:ts=4:et:



import time
from threading import Thread, RLock
from typing import Dict, Optional, List

from platypush.backend.sensor import SensorBackend
from platypush.context import get_plugin
from platypush.message.event.bluetooth import BluetoothDeviceFoundEvent, BluetoothDeviceLostEvent


class BluetoothScannerBackend(SensorBackend):
    """
    This backend periodically scans for available bluetooth devices and returns events when a devices enter or exits
    the range.

    Triggers:

        * :class:`platypush.message.event.bluetooth.BluetoothDeviceFoundEvent` when a new bluetooth device is found.
        * :class:`platypush.message.event.bluetooth.BluetoothDeviceLostEvent` when a bluetooth device is lost.

    Requires:

        * The :class:`platypush.plugins.bluetooth.BluetoothPlugin` plugin working.

    """

    def __init__(self, device_id: Optional[int] = None, scan_duration: int = 10,
                 track_devices: Optional[List[str]] = None, **kwargs):
        """
        :param device_id: Bluetooth adapter ID to use (default configured on the ``bluetooth`` plugin if None).
        :param scan_duration:  How long the scan should run (default: 10 seconds).
        :param track_devices: List of addresses of devices to actively track, even if they aren't discoverable.
        """
        super().__init__(plugin='bluetooth', plugin_args={
            'device_id': device_id,
            'duration': scan_duration,
        }, **kwargs)

        self._last_seen_devices = {}
        self._tracking_thread: Optional[Thread] = None
        self._bt_lock = RLock()
        self.track_devices = set(track_devices or [])
        self.scan_duration = scan_duration

    def _add_last_seen_device(self, dev):
        addr = dev.pop('addr')
        if addr not in self._last_seen_devices:
            self.bus.post(BluetoothDeviceFoundEvent(address=addr, **dev))
        self._last_seen_devices[addr] = {'addr': addr, **dev}

    def _remove_last_seen_device(self, addr: str):
        dev = self._last_seen_devices.get(addr)
        if not dev:
            return

        self.bus.post(BluetoothDeviceLostEvent(address=addr, **dev))
        del self._last_seen_devices[addr]

    def _addr_tracker(self, addr):
        with self._bt_lock:
            name = get_plugin('bluetooth').lookup_name(addr, timeout=self.scan_duration).name

        if name is None:
            self._remove_last_seen_device(addr)
        else:
            self._add_last_seen_device({'addr': addr, 'name': name})

    def _bt_tracker(self):
        self.logger.info('Starting Bluetooth tracker')
        while not self.should_stop():
            trackers = []
            for addr in self.track_devices:
                tracker = Thread(target=self._addr_tracker, args=(addr,))
                tracker.start()
                trackers.append(tracker)

            for tracker in trackers:
                tracker.join(timeout=self.scan_duration)

            time.sleep(self.scan_duration)

        self.logger.info('Bluetooth tracker stopped')

    def get_measurement(self):
        with self._bt_lock:
            return super().get_measurement()

    def process_data(  # lgtm [py/inheritance/signature-mismatch]
        self, data: Dict[str, dict], new_data: Optional[Dict[str, dict]] = None, **_
    ):
        for addr, dev in data.items():
            self._add_last_seen_device(dev)

        for addr, dev in self._last_seen_devices.copy().items():
            if addr not in data and addr not in self.track_devices:
                self._remove_last_seen_device(addr)

    def run(self):
        self._tracking_thread = Thread(target=self._bt_tracker)
        self._tracking_thread.start()
        super().run()

    def on_stop(self):
        super().on_stop()
        if self._tracking_thread and self._tracking_thread.is_alive():
            self.logger.info('Waiting for the Bluetooth tracking thread to stop')
            self._tracking_thread.join(timeout=self.scan_duration)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import List, Union

from platypush.plugins import Plugin, action


class SwitchPlugin(Plugin, ABC):
    """
    Abstract class for interacting with switch devices
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @action
    @abstractmethod
    def on(self, device, *args, **kwargs):
        """ Turn the device on """
        raise NotImplementedError()

    @action
    @abstractmethod
    def off(self, device, *args, **kwargs):
        """ Turn the device off """
        raise NotImplementedError()

    @action
    @abstractmethod
    def toggle(self, device, *args, **kwargs):
        """ Toggle the device status (on/off) """
        raise NotImplementedError()

    @action
    def switch_status(self, device=None) -> Union[dict, List[dict]]:
        """
        Get the status of a specified device or of all the configured devices (default).

        :param device: Filter by device name or ID.
        :return: .. schema:: switch.SwitchStatusSchema(many=True)
        """
        devices = self.switches
        if device:
            devices = [d for d in self.switches if d.get('id') == device or d.get('name') == device]
            return devices[0] if devices else []

        return devices

    @action
    def status(self, device=None, *args, **kwargs) -> Union[dict, List[dict]]:
        """
        Get the status of all the devices, or filter by device name or ID (alias for :meth:`.switch_status`).

        :param device: Filter by device name or ID.
        :return: .. schema:: switch.SwitchStatusSchema(many=True)
        """
        return self.switch_status(device)

    @property
    @abstractmethod
    def switches(self) -> List[dict]:
        """
        :return: .. schema:: switch.SwitchStatusSchema(many=True)
        """
        raise NotImplementedError()


# vim:sw=4:ts=4:et:



from marshmallow import fields
from marshmallow.schema import Schema


device_types = [
    'Hub',
    'Hub Plus',
    'Hub Mini',
    'Bot',
    'Curtain',
    'Plug',
    'Meter',
    'Humidifier',
    'Smart Fan',
    'Air Conditioner',
    'TV',
    'Light',
    'IPTV / Streamer',
    'Set Top Box',
    'DVD',
    'Fan',
    'Projector',
    'Camera',
    'Air Purifier',
    'Speaker',
    'Water Heater',
    'Vacuum Cleaner',
    'Remote',
    'Others',
]

remote_types = [
    'Air Conditioner',
    'TV',
    'Light',
    'IPTV / Streamer',
    'Set Top Box',
    'DVD',
    'Fan',
    'Projector',
    'Camera',
    'Air Purifier',
    'Speaker',
    'Water Heater',
    'Vacuum Cleaner',
    'Others',
]


class DeviceSchema(Schema):
    id = fields.String(attribute='deviceId', required=True, metadata=dict(description='Device unique ID'))
    name = fields.String(attribute='deviceName', metadata=dict(description='Device name'))
    device_type = fields.String(
        attribute='deviceType',
        metadata=dict(description=f'Default types: [{", ".join(device_types)}]')
    )
    remote_type = fields.String(
        attribute='remoteType',
        metadata=dict(description=f'Default types: [{", ".join(remote_types)}]')
    )
    hub_id = fields.String(attribute='hubDeviceId', metadata=dict(description='Parent hub device unique ID'))
    cloud_service_enabled = fields.Boolean(
        attribute='enableCloudService',
        metadata=dict(
            description='True if cloud access is enabled on this device,'
                        'False otherwise. Only cloud-enabled devices can be '
                        'controlled from the switchbot plugin.'
        )
    )
    is_calibrated = fields.Boolean(
        attribute='calibrate',
        metadata=dict(
            description='[Curtain devices only] Set to True if the device has been calibrated, False otherwise'
        )
    )
    open_direction = fields.String(
        attribute='openDirection',
        metadata=dict(
            description='[Curtain devices only] Direction where the curtains will be opened ("left" or "right")'
        )
    )
    is_virtual = fields.Boolean(
        metadata=dict(
            description='True if this is a virtual device, i.e. a device with an IR remote configuration but not '
                        'managed directly by the Switchbot bridge'
        )
    )


class DeviceStatusSchema(DeviceSchema):
    on = fields.Boolean(attribute='power', metadata=dict(description='True if the device is on, False otherwise'))
    moving = fields.Boolean(
        metadata=dict(
            description='[Curtain devices only] True if the device is moving, False otherwise'
        )
    )
    position = fields.Int(
        attribute='slidePosition', metadata=dict(
            description='[Curtain devices only] Position of the device on the curtain rail, between '
                        '0 (open) and 1 (closed)'
        )
    )
    temperature = fields.Float(
        metadata=dict(description='[Meter/humidifier/Air conditioner devices only] Temperature in Celsius')
    )
    humidity = fields.Float(metadata=dict(description='[Meter/humidifier devices only] Humidity in %'))
    fan_speed = fields.Int(
        metadata=dict(description='[Air conditioner devices only] Speed of the fan')
    )
    nebulization_efficiency = fields.Float(
        attribute='nebulizationEfficiency',
        metadata=dict(
            description='[Humidifier devices only] Nebulization efficiency in %'
        )
    )
    auto = fields.Boolean(
        metadata=dict(
            description='[Humidifier devices only] True if auto mode is on'
        )
    )
    child_lock = fields.Boolean(
        attribute='childLock',
        metadata=dict(
            description='[Humidifier devices only] True if safety lock is on'
        )
    )
    sound = fields.Boolean(
        metadata=dict(
            description='[Humidifier devices only] True if sound is muted'
        )
    )
    mode = fields.Int(
        metadata=dict(description='[Fan/Air conditioner devices only] Fan mode')
    )
    speed = fields.Float(
        metadata=dict(
            description='[Smart fan devices only] Fan speed, between 1 and 4'
        )
    )
    swinging = fields.Boolean(
        attribute='shaking',
        metadata=dict(description='[Smart fan devices only] True if the device is swinging')
    )
    swing_direction = fields.Int(
        attribute='shakeCenter',
        metadata=dict(description='[Smart fan devices only] Swing direction')
    )
    swing_range = fields.Float(
        attribute='shakeRange',
        metadata=dict(description='[Smart fan devices only] Swing range angle, between 0 and 120')
    )


class SceneSchema(Schema):
    id = fields.String(attribute='sceneId', required=True, metadata=dict(description='Scene ID'))
    name = fields.String(attribute='sceneName', metadata=dict(description='Scene name'))
